/**
 * Method for running a single process
 * @param {SubqueueKeyType} [subqueue] Unique key of the subqueue being run
 * @returns {FetchResult} Result of the process
 */
export declare type FetcherProcess<FetchResult, SubqueueKeyType = string | number | symbol> = (subqueue?: SubqueueKeyType) => Promise<FetchResult>;
/**
 * Method for running a batch fetch process
 * @param {SubqueueKeyType[]} subqueues Unique keys of the subqueues being run
 * @param {Function} earlyWrite Mechanism for unblocking subqueues as the data is available
 * @returns {FetchResult[] | Error[] | Map | void} Results in array/map format. Nothing should be returned when using the earlyWrite mechanism
 */
export declare type BatchFetcherProcess<FetchResult, SubqueueKeyType = string | number | symbol> = (subqueues: SubqueueKeyType[], earlyWrite: (subqueue: SubqueueKeyType, result: FetchResult | Error) => void) => Promise<Array<FetchResult | Error> | Map<SubqueueKeyType, FetchResult | Error> | void>;
/**
 * Configurable parameters when creating a new instance of BurstValve
 */
export interface BurstValveParams<DrainResult, SubqueueKeyType> {
    /**
     * Display name for the valve (useful for debugging exceptions)
     */
    displayName?: string;
    /**
     * Fetcher process for single concurrency process running
     */
    fetch?: FetcherProcess<DrainResult, SubqueueKeyType>;
    /**
     * Fetcher process for single concurrency on a list
     * of unique identifiers
     */
    batch?: BatchFetcherProcess<DrainResult, SubqueueKeyType>;
}
/**
 * Concurrent queue for a single (or batch) asynchronous action
 */
export declare class BurstValve<DrainResult, SubqueueKeyType = string | number | symbol> {
    /**
     * Display name for the valve
     * @type {string}
     * @readonly
     */
    readonly displayName: string;
    /**
     * Fetcher for single concurrent running of a function
     * @type {FetcherProcess | undefined}
     * @readonly
     * @private
     */
    private readonly fetcher?;
    /**
     * Fetcher for batch of unique identifiers to run once
     * @type {BatchFetcherProcess | undefined}
     * @readonly
     * @private
     */
    private readonly batchFetcher?;
    /**
     * Queue of promise callbacks
     * @type {PromiseStore[]}
     * @private
     */
    private queue?;
    /**
     * Keyed subqueues of promise callbacks
     * @type {Map}
     * @private
     */
    private subqueues;
    /**
     * Creates an instance of BurstValve with a custom fetcher
     * @param {FetcherProcess} fetcher Fetcher process for single concurrency process running
     */
    constructor(fetcher: FetcherProcess<DrainResult, SubqueueKeyType>);
    /**
     * Creates an instance of BurstValve with a custom display name and fetcher
     * @param {string} displayName Name for the valve
     * @param {FetcherProcess} fetcher Fetcher process for single concurrency process running
     */
    constructor(displayName: string, fetcher: FetcherProcess<DrainResult, SubqueueKeyType>);
    /**
     * Creates an instance of BurstValve with configurable parameters
     * @param {BurstValveParams} config Burst value configuration
     */
    constructor(config: BurstValveParams<DrainResult, SubqueueKeyType>);
    /**
     * Determines if queue (or subqueue) has an active action being taken
     * @param {SubqueueKeyType} [subqueue] Unique identifier of the subqueue to check activity.
     * @returns {Boolean} True/False indicating if queue (or subqueue) is active
     */
    isActive(subqueue?: SubqueueKeyType): boolean;
    /**
     * Leverages the current valve to only have a single running process of a function
     * @param {SubqueueKeyType} [subqueue] Unique identifier of the subqueue to fetch data for
     * @returns {DrainResult} Result of the fetch
     */
    fetch(subqueue?: SubqueueKeyType): Promise<DrainResult>;
    /**
     * Batches fetching of unique identifiers into a single process, waiting
     * for existing queues if they already exist
     * @param {SubqueueKeyType[]} subqueues List of unique identifiers to fetch at once
     * @returns {Array<DrainResult | Error>} List of fetch results or exceptions
     */
    batch(subqueues: SubqueueKeyType[]): Promise<Array<DrainResult | Error>>;
    /**
     * Same as batch, except throws any errors that are found during the fetching
     * process rather returning them. Simplifies the return array to only results
     * @param {SubqueueKeyType[]} subqueues List of unique identifiers to fetch at once
     * @returns {DrainResult[]} List of batch results
     */
    unsafeBatch(subqueues: SubqueueKeyType[]): Promise<Array<DrainResult>>;
    /**
     * Exposes results for fetching each unique identifier as the data becomes available
     * @param {SubqueueKeyType[]} subqueues List of unique identifiers to fetch at once
     * @param {Function} streamResultCallback Iterative callback for each result as it is available
     */
    stream(subqueues: SubqueueKeyType[], streamResultCallback: (subqueue: SubqueueKeyType, result: DrainResult | Error) => Promise<void>): Promise<void>;
    /**
     * Normalized runner for batch and batchUnsafe
     * @param {SubqueueKeyType[]} subqueues List of unique identifiers to fetch at once
     * @returns {Array<DrainResult | Error>} List of batch results or exceptions
     */
    private runBatch;
    /**
     * Runs the user defined batch fetcher process
     * @param {SubqueueKeyType[]} subqueues List of unique identifiers to fetch at once
     * @param {Map} [results] Optional list of shared results
     * @param {Boolean} [raiseExceptions] Indicates if errors should be thrown rather than returned
     */
    private runBatchFetcher;
    /**
     * Flushes the queue specified with the result passed
     * @param {SubqueueKeyType | undefined} subqueue Unique identifier tied to the fetch process
     * @param {DrainResult | Error} result Successful/Failed result of the fetch process
     */
    private flushResult;
}
//# sourceMappingURL=index.d.ts.map